#ifndef JEDFILE_H
#define JEDFILE_H

#pragma once

#include <atlfile.h>

#define bsf(a,b) (a)[(b)/8]|=(1<<((b)%8))
#define bcf(a,b) (a)[(b)/8]&=(0xFF^(1<<((b)%8)))

#ifndef JEDFILE_MAX
#define JEDFILE_MAX (16*1024*1024)
#endif

int DecodeJedFile(LPCTSTR filename, PUCHAR &fuses, int &bitlen, DWORD &usercode){
	CAtlFile fin;
	if(FAILED(fin.Create(filename,GENERIC_READ,FILE_SHARE_READ,OPEN_EXISTING))){
		throw _T("Could not load file.");
	}
	unsigned long long len;
	fin.GetSize(len);
	if(JEDFILE_MAX<len){
		fin.Close();
		throw _T("Too large file.");
	}
	char* buf=new char[len+1];
	DWORD loaded;
	fin.Read(buf,len, loaded);
	fin.Close();
	if(len!=loaded){
		delete[] buf;
		throw _T("File load failed.");
	}
	buf[len]=0;

	int defbit=-1, pincnt=0;
	bitlen=0;
	fuses=0;

	DEBUG_PRINTF(_T("jedfile: Loaded %dbytes.\n"),len);
	if(buf[0]!=2){
		delete[] buf;
		throw _T("Not JEDEC file.");
	}
	int i,j,k;
	for(i=1;i<len;i++){
		switch(buf[i]){
			//vf
			case 'N' :
			case 'G' :
			case 'V' :
			case 'C' :
				DEBUG_PRINTF(_T("%c"),buf[i]);
				break;
			case 'U' :
				sscanf(&buf[i+2],"%8X",&usercode);
				break;
			case 'Q' : 
				if(buf[i+1]=='P'){
					i+=2;
					pincnt=atoi(&buf[i]);
					DEBUG_PRINTF(_T("\nQP%d\n"),pincnt);
				}else if(buf[i+1]=='F'){
					i+=2;
					bitlen=atoi(&buf[i]);
					DEBUG_PRINTF(_T("\nQF%d\n"),bitlen);
				}else{
					DEBUG_PRINTF(_T("\nQ%c??\n"),buf[i+1]);
				}
				break;
			case 'F' : 
				i++;
				defbit=atoi(&buf[i]);
				DEBUG_PRINTF(_T("\njedfile: Default bit is %d.\n"),defbit);
				break;
			case 'L' : 
				if(bitlen==0 || defbit==-1){
					delete[] buf;
					delete[] fuses;
					fuses = NULL;
					throw _T("Purse error.");
				}
				if(!fuses){
					fuses=new UCHAR[(bitlen+7)/8];
					for(j=0;j<(bitlen+7)/8;j++) fuses[j]=defbit?0xFF:0;
				}
				i++;
				k=j=atoi(&buf[i]);
				while(1){
					if(buf[i]<'0' || '9'<buf[i]) break;
					i++;
				}
				for(;1;){
					switch(buf[i+1]){
						case '0' : bcf(fuses,j); j++; break;
						case '1' : bsf(fuses,j); j++; break;
						default  : break;
					}
					if(buf[i+1]==0 || buf[i+1]=='*') break;
					i++;
				}
				DEBUG_PRINTF(_T("\njedfile: L from bit%d, %dbits.\n"),k,j-k);
				break;
			//󔒕
			case '*' :
			case ' ' :
			case '\r' :
			case '\n' :
			case '\t' :
				DEBUG_PRINTF(_T("."));
				continue;
			//I[
			case 3 : 
				break;
			//s
			//default : 
				//TRACE(traceAppMsg,0,_T("\nUnrecognized char '0x%02X'.\n"),buf[i]);
		}
		//I[
		if(buf[i]==3) break;
		//f~^܂Ői߂B
		while(1){
			i++;
			if(buf[i]=='*') break;
			if(buf[i]==0  ) break;
		}
	}
	delete[] buf;
	return true;
}

#endif
