#include "stdafx.h"

#include "jtag.h"

#include "writercore.h"

//macro and functions
#define IDLE1ms 3200
#define FSMAX (1024*1024*16)
#define BITTEST(a,b) ((a)[(b)/8]&(1<<((b)%8))?1:0)

#define PIPELINED_BITS 1

unsigned CompareBits(PUCHAR a,PUCHAR b,unsigned bitlen,unsigned a_sbit=0,unsigned b_sbit=0){
	unsigned errors=0;
	while(bitlen--){
		if(BITTEST(a,a_sbit)!=BITTEST(b,b_sbit)) errors++;
		a_sbit++;
		b_sbit++;
	}
	return errors;
}

#define bsf(a,b) (a)[(b)/8]|=(1<<((b)%8))
#define bcf(a,b) (a)[(b)/8]&=(0xFF^(1<<((b)%8)))

typedef struct tag_fpgainfo{
	DWORD idcode;
	TCHAR *name; 
	DWORD preload;
	DWORD column;
	DWORD row;
} FPGAINFO;

static FPGAINFO fpgainfotable[]={
	{0x01299043,_T("LFXP2-5E"),394,638,1938},
	{0x0129B043,_T("LFXP2-17E"),730,2188,1658}
};
static FPGAINFO nomatch={0};

FPGAINFO* GetFpgaInfo(DWORD idcode){
	for(int i=0;i<sizeof(fpgainfotable)/sizeof(FPGAINFO);i++){
		if(fpgainfotable[i].idcode==idcode) return fpgainfotable+i;
	}
	return NULL;
}


int Writer_Open(){
	return JtagCableOpen();
}
void Writer_Close(){
	JtagCableClose();
}

DWORD Writer_GetIdcode(){
	DWORD idcode=0;
	UCHAR tx[16],rx[16];
	tx[0]=0x16;	//IDCODE
	JtagQueSIR(tx,0,8);
	JtagQueSDR(tx,0,32,1);
	if(false==JtagFlushQue(rx,32)) throw _T("JTAG cable error.");
	idcode=*(DWORD*)rx;
	if(!GetFpgaInfo(idcode)){
		CString errstr;
		errstr.Format(_T("Unknown IDCODE. (0x%08X)"),idcode);
		throw errstr;
	}
	return idcode;
}

int Writer_Erase(DWORD idcode, void(*cbfunc)(int,void*), void* cbptr){
	UCHAR tx[128],rx[16];
	//unsigned toc=0;
	
	tx[0]=0x1C;	//Preload
	JtagQueSIR(tx,0,8);
	for(int i=0;i<(GetFpgaInfo(idcode)->preload+7)/8;i++) tx[i]=0xFF;
	JtagQueSDR(tx,0,GetFpgaInfo(idcode)->preload);

	tx[0]=0x15;	//Enter programming mode
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);

	tx[0]=0xB2;	//Read status
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);
	tx[0]=0;
	JtagQueSDR(tx,0,8,1);
	if(false==JtagFlushQue(rx,8)) throw _T("JTAG cable error.");
	if(rx[0]&0x54) throw _T("status error.");

	cbfunc(5,cbptr);

	tx[0]=0x03;	//Erase
	JtagQueSIR(tx,0,8);

	for(int cnt=0;1;cnt++){
		JtagQueIDLE(800);
		tx[0]=0x52;
		JtagQueSIR(tx,0,8);
		tx[0]=0;
		JtagQueSDR(tx,0,1,1);
		if(false==JtagFlushQue(rx,1)) throw _T("JTAG cable error.");
		if(rx[0]&1) break;
		cbfunc(70<cnt?75:cnt+5,cbptr);
		Sleep(50);
		if(400<cnt) throw _T("Erase time out!!");
	}

	tx[0]=0x1E;	//Leave programming mode
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);
	if(false==JtagFlushQue()) throw _T("JTAG cable error.");

	cbfunc(100,cbptr);

	return true;
}

int Writer_Write(DWORD idcode, UCHAR* fuses, DWORD usercode, void(*cbfunc)(int,void*), void* cbptr, DWORD bitlen){
	UCHAR *rx = new UCHAR[512];
	UCHAR *tx = new UCHAR[512];
	unsigned wrotebits=0;

	if(bitlen-32!=GetFpgaInfo(idcode)->column*GetFpgaInfo(idcode)->row) throw _T("Bitstream length mismatch.");

	tx[0]=0x15;	//Enter programming mode
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);

	tx[0]=0xB2;	//Read Status
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);
	tx[0]=0;
	JtagQueSDR(tx,0,8,1);
	if(false==JtagFlushQue(rx,8)) throw _T("JTAG cable error.");
	DEBUG_PRINTF(_T("writercore: 0xB2->%d(expect 0) \n"),rx[0]&3);
	if(rx[0]&3) throw _T("status error.");

	tx[0]=0x21;	//InitAddress
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(8);
	while(wrotebits<(GetFpgaInfo(idcode)->column*GetFpgaInfo(idcode)->row)){
		DEBUG_PRINTF(_T("writercore: Writing row #%d...\n"),wrotebits/GetFpgaInfo(idcode)->column);
		tx[0]=0x67;	//ProgramIncrRti
		JtagQueSIR(tx,0,8);
		JtagQueSDR(fuses,wrotebits,GetFpgaInfo(idcode)->column);
		JtagQueIDLE(800);
		for(int cnt=0;1;cnt++){
			JtagQueIDLE(8);
			tx[0]=0x52;	//Program/Status
			JtagQueSIR(tx,0,8);
			tx[0]=0;
			JtagQueSDR(tx,0,1,1);
			if(false==JtagFlushQue(rx,1)) throw _T("JTAG cable error.");
			if(rx[0]&1) break;
			if(50<cnt) throw _T("Write timeout!");
		}
		if((wrotebits/GetFpgaInfo(idcode)->column%20)==0){
			cbfunc(wrotebits*95/GetFpgaInfo(idcode)->column/GetFpgaInfo(idcode)->row,cbptr);
			//this->UpdateWindow();
			//TRACE(traceAppMsg,0,_T("Wrote %d rows.\n"),wrotebits/GetFpgaInfo(v_idcode)->column);
		}
		wrotebits+=GetFpgaInfo(idcode)->column;
	}

	tx[0]=0x1A;	//program user code
	JtagQueSIR(tx,0,8);
	memcpy(tx,&usercode,4);
	JtagQueSDR(tx,0,32);
	JtagQueIDLE(IDLE1ms*10);

	tx[0]=0xB2;	//Read Status
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);
	tx[0]=0;
	JtagQueSDR(tx,0,8,1);
	if(false==JtagFlushQue(rx,8)) throw _T("JTAG cable error.");
	DEBUG_PRINTF(_T("writercore: 0xB2->%d(expect 0) \n"),rx[0]&1);
	if(rx[0]&1) throw _T("status error.");
	cbfunc(100,cbptr);
	return true;
}

int Writer_Verify(DWORD idcode, UCHAR* fuses, DWORD usercode, void(*cbfunc)(int,void*), void* cbptr, DWORD bitlen){
	UCHAR *rx = new UCHAR[512];
	UCHAR *tx = new UCHAR[512];
	DWORD v_usercode;
	DWORD v_sedcrc;
	
	if(bitlen-32!=GetFpgaInfo(idcode)->column*GetFpgaInfo(idcode)->row) throw _T("Bitstream length mismatch.");

	tx[0]=0xB2;	//Read Status
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);
	tx[0]=0;
	JtagQueSDR(tx,0,8,1);
	if(false==JtagFlushQue(rx,8)) throw _T("JTAG cable error.");
	//TRACE(traceAppMsg,0,_T("0xB2->%d \n"),rx[0]&1);
	if(rx[0]&0x80) throw _T("status error.");

	unsigned wrotebits=0,readbits=0;
	unsigned erbits=0;

	tx[0]=0x21;	//Reseet addr
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);
	tx[0]=0x6A;	//read incr rti
	JtagQueSIR(tx,0,8);
	for(int i=0; i<(GetFpgaInfo(idcode)->column+7)/8; i++) tx[i]=0x00;
	for(wrotebits=0; wrotebits<PIPELINED_BITS; wrotebits+=GetFpgaInfo(idcode)->column){
		JtagQueIDLE(8);
		JtagQueSDR(tx,0,GetFpgaInfo(idcode)->column,1);
	}
	while(readbits<GetFpgaInfo(idcode)->column*GetFpgaInfo(idcode)->row){
		DEBUG_PRINTF(_T("writercore: Verifying row #%d...\n"),readbits/GetFpgaInfo(idcode)->column);
		if(wrotebits<GetFpgaInfo(idcode)->column*GetFpgaInfo(idcode)->row){
			JtagQueIDLE(8);
			JtagQueSDR(tx,0,GetFpgaInfo(idcode)->column,1);
		}
		if(false==JtagFlushQue(rx,GetFpgaInfo(idcode)->column,0)) throw _T("JTAG cable error.");
		erbits+=CompareBits(fuses,rx,GetFpgaInfo(idcode)->column,readbits);
		if((readbits/GetFpgaInfo(idcode)->column%20)==0){
			cbfunc(readbits*95/GetFpgaInfo(idcode)->column/GetFpgaInfo(idcode)->row,cbptr);
			//TRACE(traceAppMsg,0,_T("Read %d rows.\n"),readbits/GetFpgaInfo(idcode)->column);
		}
		#if 0
		if((readbits/GetFpgaInfo(idcode)->column)<10){
			TRACE(traceAppMsg,0,_T("row%4d: "),readbits/GetFpgaInfo(v_idcode)->column+1);
			for(int i=79;0<=i;i--) TRACE(traceAppMsg,0,_T("%02X "),rx[i]);
			TRACE(traceAppMsg,0,_T("\n"));
		}
		#endif
		readbits+=GetFpgaInfo(idcode)->column;
		wrotebits+=GetFpgaInfo(idcode)->column;
	}

	tx[0]=0x17;	//Read Usercode
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);
	tx[0]=0xFF; tx[1]=0xFF; tx[2]=0xFF; tx[3]=0xFF;
	JtagQueSDR(tx,0,32,1);
	if(false==JtagFlushQue(rx,32)) throw _T("JTAG cable error.");
	v_usercode=*(DWORD*)rx;

	tx[0]=0x45;	//program sed crc
	JtagQueSIR(tx,0,8);
	JtagQueSDR(fuses,GetFpgaInfo(idcode)->column*GetFpgaInfo(idcode)->row,32);
	JtagQueIDLE(IDLE1ms);

	tx[0]=0x44;	//read sed crc
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(8);
	tx[0]=0; tx[1]=0; tx[2]=0; tx[3]=0; 
	JtagQueSDR(tx,0,32,1);
	if(false==JtagFlushQue(rx,32)) throw _T("JTAG cable error.");
	v_sedcrc=*(DWORD*)rx;

	tx[0]=0xB2;	//Read Status
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);
	tx[0]=0;
	JtagQueSDR(tx,0,8,1);
	if(false==JtagFlushQue(rx,8)) throw _T("JTAG cable error.");
	DEBUG_PRINTF(_T("writercore: 0xB2->%d(expect 0) \n"),rx[0]&1);
	if(rx[0]&1) throw _T("status error.");
		
	if(erbits||(usercode!=v_usercode)||(CompareBits(fuses,(PUCHAR)&v_sedcrc,32,GetFpgaInfo(idcode)->column*GetFpgaInfo(idcode)->row))){
		CString tmp,msg;
		msg=_T("Verify Errors!!\n");
		tmp.Format(_T("Config data mismatch : %d / %d bits\n"),erbits,GetFpgaInfo(idcode)->column*GetFpgaInfo(idcode)->row);
		msg+=tmp;
		tmp.Format(_T("   Usercode mismatch : %d / %d bits\n"),CompareBits((PUCHAR)&usercode,(PUCHAR)&v_usercode,32),32);
		msg+=tmp;
		tmp.Format(_T("    SED_CRC mismatch : %d / %d bits\n"),CompareBits(fuses,(PUCHAR)&v_sedcrc,32,GetFpgaInfo(idcode)->column*GetFpgaInfo(idcode)->row),32);
		msg+=tmp;
		MessageBox(NULL,msg,NULL,NULL);
		throw _T("Verify error.");
	}


	tx[0]=0x2F;	//program done bit
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);
	if(false==JtagFlushQue()) throw _T("JTAG cable error.");
	Sleep(200);
		
	tx[0]=0xB2;	//Read Status
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);
	tx[0]=0;
	JtagQueSDR(tx,0,8,1);
	if(false==JtagFlushQue(rx,8)) throw _T("JTAG cable error.");
	DEBUG_PRINTF(_T("writercore: 0xB2->%d(expect 2) \n"),rx[0]&3);
	//if(rx[0]&1) throw _T("status error.");

	tx[0]=0x1E;	//Leave programming mode
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);
	tx[0]=0xFF;	//Bypass ...to enter usermode?
	JtagQueSIR(tx,0,8);
	JtagQueIDLE(IDLE1ms);
	if(false==JtagFlushQue()) throw _T("JTAG cable error.");
	cbfunc(100,cbptr);

	return true;
}
